"""Even-higher-level APIs over PySNMP's high-level APIs"""

import logging
import os
import threading
from collections import defaultdict
from ipaddress import ip_address
from typing import Any, Sequence, Tuple, Union

from pyasn1.type import univ
from pysnmp.hlapi.asyncio import (
    CommunityData,
    ContextData,
    ObjectIdentity,
    ObjectType,
    SnmpEngine,
    Udp6TransportTarget,
    UdpTransportTarget,
    bulkCmd,
    getCmd,
    nextCmd,
)
from pysnmp.proto import errind
from pysnmp.proto.rfc1905 import EndOfMibView, NoSuchInstance, NoSuchObject, errorStatus
from pysnmp.smi import builder, view
from pysnmp.smi.error import MibNotFoundError as PysnmpMibNotFoundError

from zino.config.models import PollDevice
from zino.oid import OID
from zino.snmp.base import (
    EndOfMibViewError,
    ErrorIndication,
    ErrorStatus,
    Identifier,
    MibNotFoundError,
    MibObject,
    NoSuchInstanceError,
    NoSuchNameError,
    NoSuchObjectError,
    SNMPVarBind,
    SparseWalkResponse,
)

_log = logging.getLogger(__name__)

# keep track of variables that need to be local to the current thread
_local = threading.local()

MIB_SOURCE_DIR = os.path.join(os.path.dirname(__file__), "mibdumps")


def _get_engine():
    if not getattr(_local, "snmp_engine", None):
        _local.snmp_engine = get_new_snmp_engine()
    return _local.snmp_engine


def init_backend():
    """Initializes the PySNMP backend"""
    _get_engine()


def get_new_snmp_engine() -> SnmpEngine:
    """Returns a new SnmpEngine object with Zino's directory of MIB modules loaded.

    The SNMP engine is a PySNMP specific object.  This may be useful when working directly with PySNMP, but it is not
    part of the high-level API that an SNMP back-end module needs to export.
    """
    snmp_engine = SnmpEngine()
    mib_builder = snmp_engine.getMibBuilder()
    mib_builder.addMibSources(builder.DirMibSource(MIB_SOURCE_DIR))
    mib_builder.loadModules()
    return snmp_engine


PySNMPVarBind = tuple[ObjectIdentity, ObjectType]
SupportedTypes = Union[univ.Integer, univ.OctetString, ObjectIdentity, ObjectType]


class SNMP:
    """Represents an SNMP management session for a single device"""

    NON_REPEATERS = 0

    def __init__(self, device: PollDevice):
        self.device = device

    # SNMP objects are expected to provide a context manager interface for allocating and releasing low-level
    # resources.  There are no low-level resources to manage in PySNMP, so these next two methods are empty:

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return

    async def get(self, *oid: str) -> MibObject:
        """SNMP-GETs the given oid
        Example usage:
            get("SNMPv2-MIB", "sysUpTime", 0)
            get("1.3.6.1.2.1.1.3.0")

        :param oid: Values for defining an OID. For detailed use see
            https://github.com/pysnmp/pysnmp/blob/bc1fb3c39764f36c1b7c9551b52ef8246b9aea7c/pysnmp/smi/rfc1902.py#L35-L49
        :return: A MibObject representing the resulting MIB variable
        """
        query = self._oid_to_object_type(*oid)
        try:
            error_indication, error_status, error_index, var_binds = await getCmd(
                _get_engine(),
                self.community_data,
                self.udp_transport_target,
                ContextData(),
                query,
            )
        except PysnmpMibNotFoundError as error:
            raise MibNotFoundError(error)
        self._raise_errors(error_indication, error_status, error_index, query)
        result = var_binds[0]
        self._raise_varbind_errors(result)
        return self._object_type_to_mib_object(result)

    async def get2(self, *variables: Sequence[Union[str, int]]) -> list[SNMPVarBind]:
        """SNMP-GETs multiple variables

        Example usage:
            get2(("SNMPv2-MIB", "sysUpTime", 0), ("SNMPv2-MIB", "sysDescr", 0))
            get2(("1.3.6.1.2.1.1.3.0",))

        :param variables: Values for defining OIDs. For detailed use see
            https://github.com/pysnmp/pysnmp/blob/bc1fb3c39764f36c1b7c9551b52ef8246b9aea7c/pysnmp/smi/rfc1902.py#L35-L49
        :return: A list of MibObject instances representing the resulting MIB variables
        """
        query = [self._oid_to_object_type(*var) for var in variables]
        try:
            error_indication, error_status, error_index, var_binds = await getCmd(
                _get_engine(),
                self.community_data,
                self.udp_transport_target,
                ContextData(),
                *query,
            )
        except PysnmpMibNotFoundError as error:
            raise MibNotFoundError(error)
        self._raise_errors(error_indication, error_status, error_index, query)
        return [_convert_varbind(*v) for v in var_binds]

    def _raise_errors(
        self,
        error_indication: Union[str, errind.ErrorIndication],
        error_status: str,
        error_index: int,
        *query: ObjectType,
    ):
        """Raises a relevant exception if an error has occurred"""
        # Local errors (timeout, config errors etc)
        if error_indication:
            if isinstance(error_indication, errind.RequestTimedOut):
                raise TimeoutError(str(error_indication))
            else:
                raise ErrorIndication(str(error_indication))

        # Remote errors from SNMP entity.
        # if nonzero error_status, error_index point will point to the ariable-binding in query that caused the error.
        if error_status:
            error_object = self._object_type_to_mib_object(query[error_index - 1])
            error_name = errorStatus.getNamedValues()[int(error_status)]
            if error_name == "noSuchName":
                raise NoSuchNameError(f"Could not find object at {error_object.oid}")
            else:
                raise ErrorStatus(f"SNMP operation failed with error {error_name} for {error_object.oid}")

    def _raise_varbind_errors(self, object_type: ObjectType):
        """Raises a relevant exception if an error has occurred in a varbind"""
        oid = OID(str(object_type[0]))
        value = object_type[1]
        if isinstance(value, NoSuchObject):
            raise NoSuchObjectError(f"Could not find object at {oid}")
        if isinstance(value, NoSuchInstance):
            raise NoSuchInstanceError(f"Could not find instance at {oid}")
        if isinstance(value, EndOfMibView):
            raise EndOfMibViewError("Reached end of MIB view")

    async def getnext(self, *oid: str) -> MibObject:
        """SNMP-GETNEXTs the given oid
        Example usage:
            getnext("SNMPv2-MIB", "sysUpTime")
            getnext("1.3.6.1.2.1.1.3")

        :param oid: Values for defining an OID. For detailed use see
            https://github.com/pysnmp/pysnmp/blob/bc1fb3c39764f36c1b7c9551b52ef8246b9aea7c/pysnmp/smi/rfc1902.py#L35-L49
        :return: A MibObject representing the resulting MIB variable
        """
        query = self._oid_to_object_type(*oid)
        object_type = await self._getnext(query)
        return self._object_type_to_mib_object(object_type)

    async def _getnext(self, object_type: ObjectType) -> ObjectType:
        """SNMP-GETNEXTs the given object_type

        :param object_type: An ObjectType representing the object you want to query
        :return: An ObjectType representing the resulting MIB variable
        """
        try:
            error_indication, error_status, error_index, var_binds = await nextCmd(
                _get_engine(),
                self.community_data,
                self.udp_transport_target,
                ContextData(),
                object_type,
            )
        except PysnmpMibNotFoundError as error:
            raise MibNotFoundError(error)
        self._raise_errors(error_indication, error_status, error_index, object_type)
        # var_binds should be a sequence of sequences with one inner sequence that contains the result.
        return var_binds[0][0]

    async def getnext2(self, *variables: Sequence[str]) -> list[SNMPVarBind]:
        """Dispatches a GET-NEXT query for the given variables and returns a result where the fetched variables are
        identified symbolically.

        Example usage:
        >>> await s.getnext2(("IF-MIB", "ifName", "1"), ("IF-MIB", "ifAlias", "1"])
        [(Identifier("IF-MIB", "ifName", OID('.2')), "Gi2/3"),
         (Identifier("IF-MIB", "ifName", OID('.2')), "Uplink to somewhere")]
        >>>

        :param variables: Values for defining OIDs. For detailed use see
        https://github.com/pysnmp/pysnmp/blob/bc1fb3c39764f36c1b7c9551b52ef8246b9aea7c/pysnmp/smi/rfc1902.py#L35-L49
        :return: A sequence of MibObject instances representing the resulting MIB variables
        """
        query = [self._oid_to_object_type(*var) for var in variables]
        response = await self._getnext2(*query)
        return [_convert_varbind(*r) for r in response]

    async def _getnext2(self, *variables: ObjectType) -> list[ObjectType]:
        """SNMP-GETNEXTs the given variables

        :param variables: An sequence of ObjectTypes representing the objects you want to query
        :return: A sequence of ObjectTypes representing the resulting MIB variables
        """
        try:
            error_indication, error_status, error_index, var_bind_table = await nextCmd(
                _get_engine(),
                self.community_data,
                self.udp_transport_target,
                ContextData(),
                *variables,
            )
        except PysnmpMibNotFoundError as error:
            raise MibNotFoundError(error)
        self._raise_errors(error_indication, error_status, error_index, *variables)
        # The table should contain only one set of results for our query
        return var_bind_table[0]

    async def walk(self, *oid: str) -> list[MibObject]:
        """Uses SNMP-GETNEXT calls to get all objects in the subtree with oid as root
        Example usage:
            walk("IF-MIB", "ifName")
            walk("1.3.6.1.2.1.31.1.1.1.1")

        :param oid: Values for defining an OID. For detailed use see
            https://github.com/pysnmp/pysnmp/blob/bc1fb3c39764f36c1b7c9551b52ef8246b9aea7c/pysnmp/smi/rfc1902.py#L35-L49
        :return: A list of MibObjects representing the resulting MIB variables
        """
        results = []
        current_object = self._oid_to_object_type(*oid)
        self._resolve_object(current_object)
        original_oid = OID(str(current_object[0]))
        while True:
            current_object = await self._getnext(current_object)
            if not original_oid.is_a_prefix_of(str(current_object[0])):
                break
            mib_object = self._object_type_to_mib_object(current_object)
            results.append(mib_object)
        return results

    async def getbulk(self, *oid: str, max_repetitions: int = 1) -> list[MibObject]:
        """SNMP-BULKs the given oid
        Example usage:
            getbulk("IF-MIB", "ifName", max_repetitions=5)
            getbulk("1.3.6.1.2.1.31.1.1.1.1")

        :param oid: Values for defining an OID. For detailed use see
            https://github.com/pysnmp/pysnmp/blob/bc1fb3c39764f36c1b7c9551b52ef8246b9aea7c/pysnmp/smi/rfc1902.py#L35-L49
        :param max_repetitions: Max amount of MIB objects to retrieve
        :return: A list of MibObjects representing the resulting MIB variables
        """
        oid_object = self._oid_to_object_type(*oid)
        objecttypes = await self._getbulk(oid_object, max_repetitions)
        results = []
        for objecttype in objecttypes:
            mibobject = self._object_type_to_mib_object(objecttype)
            results.append(mibobject)
        return results

    async def _getbulk(self, object_type: ObjectType, max_repetitions: int) -> list[ObjectType]:
        """SNMP-BULKs the given `oid_object`"""
        try:
            error_indication, error_status, error_index, var_binds = await bulkCmd(
                _get_engine(),
                self.community_data,
                self.udp_transport_target,
                ContextData(),
                self.NON_REPEATERS,
                max_repetitions,
                object_type,
            )
        except PysnmpMibNotFoundError as error:
            raise MibNotFoundError(error)
        self._raise_errors(error_indication, error_status, error_index, object_type)
        return var_binds[0]

    async def getbulk2(self, *variables: Sequence[str], max_repetitions: int = 10) -> list[list[SNMPVarBind]]:
        """Issues a GET-BULK request for a set of multiple variables, returning the response in a slightly different
         format than getbulk.

        Example usage:
            >>> snmp = SNMP(...)
            >>> snmp.getbulk2(("IF-MIB", "ifName"), ("IF-MIB", "ifAlias"), max_repetitions=5)
            [[(Identifier(mib='IF-MIB', object='ifAlias', index=OID('.1')), 'Uplink'),
              (Identifier(mib='IF-MIB', object='ifDescr', index=OID('.1')), 'GigabitEthernet0/1')],
             [(Identifier(mib='IF-MIB', object='ifAlias', index=OID('.2')), 'example-sw.example.org'),
              (Identifier(mib='IF-MIB', object='ifDescr', index=OID('.2')), 'GigabitEthernet0/2')]]

        :param variables: Variables to fetch, either as OIDs or symbolic names
        :param max_repetitions: Max amount of MIB objects to retrieve
        :return: A sequence of two-tuples that represent the response varbinds
        """
        oid_objects = [self._oid_to_object_type(*var) for var in variables]
        var_bind_table = await self._getbulk2(*oid_objects, max_repetitions=max_repetitions)

        return [[_convert_varbind(i, v) for i, v in var_binds] for var_binds in var_bind_table]

    async def _getbulk2(self, *variables: Sequence[ObjectType], max_repetitions: int) -> list[list[PySNMPVarBind]]:
        """Issues a GET-BULK request for one or more variables, returning the raw var bind table from PySNMP"""
        try:
            error_indication, error_status, error_index, var_bind_table = await bulkCmd(
                _get_engine(),
                self.community_data,
                self.udp_transport_target,
                ContextData(),
                self.NON_REPEATERS,
                max_repetitions,
                *variables,
            )
        except PysnmpMibNotFoundError as error:
            raise MibNotFoundError(error)
        self._raise_errors(error_indication, error_status, error_index, *variables)
        return var_bind_table

    async def bulkwalk(self, *oid: str, max_repetitions: int = 10) -> list[MibObject]:
        """Uses SNMP-BULK calls to get all objects in the subtree with oid as root
        Example usage:
            bulkwalk("IF-MIB", "ifName", max_repetitions=5)
            bulkwalk("1.3.6.1.2.1.31.1.1.1.1")

        :param oid: Values for defining an OID. For detailed use see
            https://github.com/pysnmp/pysnmp/blob/bc1fb3c39764f36c1b7c9551b52ef8246b9aea7c/pysnmp/smi/rfc1902.py#L35-L49
        :param max_repetitions: Max amount of MIB objects to retrieve per SNMP-BULK call
        :return: A list of MibObjects representing the resulting MIB variables
        """
        results = []
        query_object = self._oid_to_object_type(*oid)
        self._resolve_object(query_object)
        start_oid = OID(str(query_object[0]))
        while True:
            response = await self._getbulk(query_object, max_repetitions)
            if not response:
                break
            for result in response:
                if not start_oid.is_a_prefix_of(str(result[0])):
                    return results
                query_object = result
                mib_object = self._object_type_to_mib_object(result)
                results.append(mib_object)
        return results

    async def sparsewalk(self, *variables: Sequence[str], max_repetitions: int = 10) -> SparseWalkResponse:
        """Bulkwalks and returns a "sparse" table.

        A sparse walk is just a walk operation that returns selected columns of table (or, from multiple tables that
        augment each other).

        Example usage:
            >>> snmp = SNMP(...)
            >>> snmp.sparsewalk(("IF-MIB", "ifName"), ("IF-MIB", "ifAlias"))
            {OID('.1'): {"ifname": "1", "ifAlias": "uplink"},
             OID('.2'): {"ifName": "2", "ifAlias": "next-sw.example.org"}}
        """
        query_objects = [self._oid_to_object_type(*var) for var in variables]
        [self._resolve_object(obj) for obj in query_objects]

        roots = [OID(o[0]) for o in query_objects]  # used to determine which responses are in scope
        results: dict[OID, dict[str, Any]] = defaultdict(dict)

        def _var_bind_is_in_scope(var: PySNMPVarBind) -> bool:
            oid = OID(var[0])
            return any(root.is_a_prefix_of(oid) for root in roots)

        while True:
            var_bind_table = await self._getbulk2(*query_objects, max_repetitions=max_repetitions)
            if not var_bind_table:
                break

            # Integrate response values into result dict
            for var_binds in var_bind_table:
                for var_bind in var_binds:
                    if _var_bind_is_in_scope(var_bind):
                        ident, value = _convert_varbind(*var_bind)
                        results[ident.index][ident.object] = value

            # Build next set of query objects from last result row
            query_objects = [v for v in var_binds if _var_bind_is_in_scope(v)]
            if not query_objects:
                break  # Nothing left to query
        return dict(results)

    def is_in_scope(self, entry: MibObject, oid: Tuple[str, str]):
        """Returns if the given MibObject is within the subtree defined by the given OID"""
        object_type = self._oid_to_object_type(*oid)
        self._resolve_object(object_type=object_type)
        root = OID(object_type[0])
        return root.is_a_prefix_of(other=entry.oid)

    async def subtree_is_supported(self, *oid: str) -> bool:
        """Returns if the device has an entry for at least one object within the subtree of the given OID"""
        entry = await self.getnext(*oid)
        if entry and self.is_in_scope(entry=entry, oid=oid):
            return True
        return False

    @staticmethod
    def _object_type_to_mib_object(object_type: ObjectType) -> MibObject:
        oid = OID(str(object_type[0]))
        value = mib_value_to_python(object_type[1])
        return MibObject(oid, value)

    @classmethod
    def _resolve_object(cls, object_type: ObjectType):
        """Raises MibNotFoundError if oid in `object` can not be found"""
        try:
            engine = _get_engine()
            controller = engine.getUserContext("mibViewController")
            if not controller:
                controller = view.MibViewController(engine.getMibBuilder())
            object_type.resolveWithMib(controller)
        except PysnmpMibNotFoundError as error:
            raise MibNotFoundError(error)

    @classmethod
    def _oid_to_object_type(cls, *oid: str) -> ObjectType:
        return ObjectType(ObjectIdentity(*oid))

    @property
    def mp_model(self) -> int:
        """Returns the preferred SNMP version of this device as a PySNMP mpModel value"""
        return 1 if self.device.snmpversion == "v2c" else 0

    @property
    def community_data(self) -> CommunityData:
        return CommunityData(self.device.community, mpModel=self.mp_model)

    @property
    def udp_transport_target(self) -> Union[UdpTransportTarget, Udp6TransportTarget]:
        assert self.device.address.version in (4, 6)
        target = UdpTransportTarget if self.device.address.version == 4 else Udp6TransportTarget
        return target(
            (str(self.device.address), self.device.port), timeout=self.device.timeout, retries=self.device.retries
        )


def _convert_varbind(ident: ObjectIdentity, value: ObjectType) -> SNMPVarBind:
    """Converts a PySNMP varbind pair to an Identifier/value pair"""
    mib, obj, indices = ident.getMibSymbol()
    value = mib_value_to_python(value)

    prefix = SNMP._oid_to_object_type(mib, obj)
    SNMP._resolve_object(prefix)
    prefix = OID(prefix[0])
    row_index = OID(ident).strip_prefix(prefix)

    return Identifier(mib, obj, row_index), value


def mib_value_to_python(value: SupportedTypes) -> Union[str, int, OID]:
    """Translates various PySNMP mib value objects to plainer Python objects, such as strings, integers or OIDs"""
    if isinstance(value, univ.Integer):
        value = int(value) if not value.namedValues else value.prettyPrint()
    elif isinstance(value, univ.OctetString):
        if type(value).__name__ in ("InetAddress", "IpAddress"):
            value = ip_address(bytes(value))
        else:
            value = str(value)
    elif isinstance(value, (ObjectIdentity, univ.ObjectIdentifier)):
        value = OID(str(value))
    else:
        raise ValueError(f"Could not convert unknown type {type(value)}")
    return value



import logging

import pytest

from zino.config.models import PollDevice
from zino.events import AlarmEvent
from zino.state import ZinoState
from zino.tasks.juniperalarmtask import JuniperAlarmTask


class TestJuniperalarmTask:
    async def test_task_runs_without_errors(self, juniper_alarm_task):
        assert (await juniper_alarm_task.run()) is None

    async def test_task_does_nothing_for_non_juniper_device(self, juniper_alarm_task):
        task = juniper_alarm_task
        device_state = task.state.devices.get(device_name=task.device.name)
        device_state.enterprise_id = 11

        await task.run()

        assert device_state.alarms is None

    async def test_task_does_nothing_for_no_result(self, caplog, snmp_test_port):
        device = PollDevice(
            name="buick.lab.example.org",
            address="127.0.0.1",
            community="cisco-bgp",  # Use some non juniper router
            port=snmp_test_port,
        )
        state = ZinoState()
        task = JuniperAlarmTask(device, state)
        device_state = task.state.devices.get(device_name=task.device.name)
        device_state.enterprise_id = 2636

        with caplog.at_level(logging.DEBUG):
            await task.run()

        assert device_state.alarms is None
        assert (
            f"Device {task.device.name} returns alarm count not of type int. Yellow alarm count: type <class 'str'>. Red alarm count: type <class 'str'>."
            not in caplog.text
        )

    async def test_task_logs_error_for_non_int_result(self, caplog, snmp_test_port):
        device = PollDevice(
            name="buick.lab.example.org",
            address="127.0.0.1",
            community="juniper-alarm-wrong",
            port=snmp_test_port,
        )
        state = ZinoState()
        task = JuniperAlarmTask(device, state)
        device_state = task.state.devices.get(device_name=task.device.name)
        device_state.enterprise_id = 2636
        with caplog.at_level(logging.DEBUG):
            await task.run()

        assert device_state.alarms is None
        assert (
            f"Device {task.device.name} returns alarm count not of type int. Yellow alarm count: type <class 'int'>. Red alarm count: type <class 'str'>."
            in caplog.text
        ) or (
            f"Device {task.device.name} returns alarm count not of type int. Yellow alarm count: type <class 'int'>. Red alarm count: type <class 'bytes'>."
            in caplog.text
        )
        assert ("Yellow alarm count: value 0. Red alarm count: value 'buick'." in caplog.text) or (
            "Yellow alarm count: value 0. Red alarm count: value b'buick'." in caplog.text
        )

    async def test_task_saves_alarm_count_in_device_state(self, juniper_alarm_task):
        task = juniper_alarm_task
        device_state = task.state.devices.get(device_name=task.device.name)
        device_state.enterprise_id = 2636

        await task.run()

        assert device_state.alarms
        assert device_state.alarms["yellow"] == 1
        assert device_state.alarms["red"] == 2

    async def test_task_overrides_alarm_count_in_device_state(self, juniper_alarm_task):
        task = juniper_alarm_task
        device_state = task.state.devices.get(device_name=task.device.name)
        device_state.enterprise_id = 2636
        device_state.alarms = {
            "yellow": 2,
            "red": 3,
        }

        await task.run()

        assert device_state.alarms["yellow"] == 1
        assert device_state.alarms["red"] == 2

    async def test_task_creates_both_alarm_events_on_both_counts_changed(self, juniper_alarm_task):
        task = juniper_alarm_task
        device_state = task.state.devices.get(device_name=task.device.name)
        device_state.enterprise_id = 2636

        await task.run()

        yellow_event = task.state.events.get(device_name=task.device.name, subindex="yellow", event_class=AlarmEvent)
        red_event = task.state.events.get(device_name=task.device.name, subindex="red", event_class=AlarmEvent)

        assert yellow_event
        assert red_event
        assert yellow_event.alarm_count == 1
        assert red_event.alarm_count == 2

    async def test_task_creates_one_alarm_event_on_one_count_changed(self, juniper_alarm_task):
        task = juniper_alarm_task
        device_state = task.state.devices.get(device_name=task.device.name)
        device_state.enterprise_id = 2636
        device_state.alarms = {
            "yellow": 1,
            "red": 0,
        }

        await task.run()

        yellow_event = task.state.events.get(device_name=task.device.name, subindex="yellow", event_class=AlarmEvent)
        red_event = task.state.events.get(device_name=task.device.name, subindex="red", event_class=AlarmEvent)

        assert not yellow_event
        assert red_event
        assert red_event.alarm_count == 2

    async def test_task_creates_event_with_correct_log_on_initial_run(self, juniper_alarm_task):
        task = juniper_alarm_task
        device_state = task.state.devices.get(device_name=task.device.name)
        device_state.enterprise_id = 2636
        device_state.alarms = {
            "yellow": 1,
            "red": 0,
        }

        await task.run()

        red_event = task.state.events.get(device_name=task.device.name, subindex="red", event_class=AlarmEvent)

        assert red_event.log
        log_messages = [log_entry.message for log_entry in red_event.log]
        assert f"{juniper_alarm_task.device.name} red alarms went from 0 to 2" in log_messages

    async def test_task_updates_alarm_events(self, juniper_alarm_task):
        task = juniper_alarm_task
        device_state = task.state.devices.get(device_name=task.device.name)
        device_state.enterprise_id = 2636
        yellow_event = task.state.events.get_or_create_event(
            device_name=task.device.name, subindex="yellow", event_class=AlarmEvent
        )
        yellow_event.alarm_type = "yellow"
        yellow_event.alarm_count = 2
        task.state.events.commit(yellow_event)
        red_event = task.state.events.get_or_create_event(
            device_name=task.device.name, subindex="red", event_class=AlarmEvent
        )
        red_event.alarm_type = "red"
        red_event.alarm_count = 3
        task.state.events.commit(red_event)

        await task.run()

        new_yellow_event, new_red_event = task.state.events[yellow_event.id], task.state.events[red_event.id]
        assert new_yellow_event.alarm_count == 1
        assert new_red_event.alarm_count == 2

    async def test_task_does_not_create_alarm_events_on_unchanged_alarm_count(self, juniper_alarm_task):
        task = juniper_alarm_task
        device_state = task.state.devices.get(device_name=task.device.name)
        device_state.enterprise_id = 2636
        device_state.alarms = {
            "yellow": 1,
            "red": 2,
        }

        await task.run()

        yellow_event = task.state.events.get(device_name=task.device.name, subindex="yellow", event_class=AlarmEvent)
        red_event = task.state.events.get(device_name=task.device.name, subindex="red", event_class=AlarmEvent)

        assert not yellow_event
        assert not red_event

    async def test_task_does_not_create_alarm_events_on_alarm_count_zero_on_first_run(self, snmp_test_port):
        device = PollDevice(
            name="buick.lab.example.org",
            address="127.0.0.1",
            community="juniper-alarm-zero",
            port=snmp_test_port,
        )
        state = ZinoState()
        task = JuniperAlarmTask(device, state)
        device_state = task.state.devices.get(device_name=task.device.name)
        device_state.enterprise_id = 2636
        await task.run()

        yellow_event = task.state.events.get(device_name=task.device.name, subindex="yellow", event_class=AlarmEvent)
        red_event = task.state.events.get(device_name=task.device.name, subindex="red", event_class=AlarmEvent)

        assert not yellow_event
        assert not red_event


@pytest.fixture()
def juniper_alarm_task(snmpsim, snmp_test_port):
    device = PollDevice(
        name="buick.lab.example.org",
        address="127.0.0.1",
        community="juniper-alarm",
        port=snmp_test_port,
    )
    state = ZinoState()
    task = JuniperAlarmTask(device, state)
    yield task



#
# PySNMP MIB module BGP4-MIB (http://snmplabs.com/pysmi)
# ASN.1 source file://./BGP4-MIB.my
# Produced by pysmi-1.1.10 at Wed Aug 23 14:18:16 2023
# On host Johanna-England-Sikt platform Linux version 6.2.0-26-generic by user johanna
# Using Python version 3.10.12 (main, Jun 11 2023, 05:26:28) [GCC 11.4.0]
#
OctetString, Integer, ObjectIdentifier = mibBuilder.importSymbols("ASN1", "OctetString", "Integer", "ObjectIdentifier")
NamedValues, = mibBuilder.importSymbols("ASN1-ENUMERATION", "NamedValues")
ConstraintsUnion, ConstraintsIntersection, SingleValueConstraint, ValueRangeConstraint, ValueSizeConstraint = mibBuilder.importSymbols("ASN1-REFINEMENT", "ConstraintsUnion", "ConstraintsIntersection", "SingleValueConstraint", "ValueRangeConstraint", "ValueSizeConstraint")
ModuleCompliance, NotificationGroup, ObjectGroup = mibBuilder.importSymbols("SNMPv2-CONF", "ModuleCompliance", "NotificationGroup", "ObjectGroup")
ModuleIdentity, iso, NotificationType, Counter32, Bits, MibScalar, MibTable, MibTableRow, MibTableColumn, IpAddress, mib_2, MibIdentifier, Unsigned32, Gauge32, TimeTicks, Integer32, Counter64, ObjectIdentity = mibBuilder.importSymbols("SNMPv2-SMI", "ModuleIdentity", "iso", "NotificationType", "Counter32", "Bits", "MibScalar", "MibTable", "MibTableRow", "MibTableColumn", "IpAddress", "mib-2", "MibIdentifier", "Unsigned32", "Gauge32", "TimeTicks", "Integer32", "Counter64", "ObjectIdentity")
TextualConvention, DisplayString = mibBuilder.importSymbols("SNMPv2-TC", "TextualConvention", "DisplayString")
bgp = ModuleIdentity((1, 3, 6, 1, 2, 1, 15))
bgp.setRevisions(('2006-01-11 00:00', '1994-05-05 00:00', '1991-10-26 18:39',))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    if mibBuilder.loadTexts: bgp.setRevisionsDescriptions(("Changes from RFC 1657: 1) Fixed the definitions of the notifications to make them equivalent to their initial definition in RFC 1269. 2) Added compliance and conformance info. 3) Updated information for the values of bgpPeerNegotiatedVersion, bgp4PathAttrLocalPref, bgp4PathAttrCalcLocalPref, bgp4PathAttrMultiExitDisc, bgp4PathAttrASPathSegement. 4) Added additional clarification comments where needed. 5) Noted where objects do not fully reflect the protocol as Known Issues. 6) Updated the DESCRIPTION for the bgp4PathAttrAtomicAggregate object. 7) The following objects have had their DESCRIPTION clause modified to remove the text that suggested (using 'should' verb) initializing the counter to zero on a transition to the established state: bgpPeerInUpdates, bgpPeerOutUpdates, bgpPeerInTotalMessages, bgpPeerOutTotalMessages Those implementations that still do this are still compliant with this new wording. Applications should not assume counters have started at zero. Published as RFC 4273.", 'Translated to SMIv2 and published as RFC 1657.', 'Initial version, published as RFC 1269.',))
if mibBuilder.loadTexts: bgp.setLastUpdated('200601110000Z')
if mibBuilder.loadTexts: bgp.setOrganization('IETF IDR Working Group')
if mibBuilder.loadTexts: bgp.setContactInfo('E-mail: idr@ietf.org Jeffrey Haas, Susan Hares (Editors) NextHop Technologies 825 Victors Way Suite 100 Ann Arbor, MI 48108-2738 Tel: +1 734 222-1600 Fax: +1 734 222-1602 E-mail: jhaas@nexthop.com skh@nexthop.com')
if mibBuilder.loadTexts: bgp.setDescription('The MIB module for the BGP-4 protocol. Copyright (C) The Internet Society (2006). This version of this MIB module is part of RFC 4273; see the RFC itself for full legal notices.')
bgpVersion = MibScalar((1, 3, 6, 1, 2, 1, 15, 1), OctetString().subtype(subtypeSpec=ValueSizeConstraint(1, 255))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpVersion.setReference('RFC 4271, Section 4.2.')
if mibBuilder.loadTexts: bgpVersion.setStatus('current')
if mibBuilder.loadTexts: bgpVersion.setDescription('Vector of supported BGP protocol version numbers. Each peer negotiates the version from this vector. Versions are identified via the string of bits contained within this object. The first octet contains bits 0 to 7, the second octet contains bits 8 to 15, and so on, with the most significant bit referring to the lowest bit number in the octet (e.g., the MSB of the first octet refers to bit 0). If a bit, i, is present and set, then the version (i+1) of the BGP is supported.')
bgpLocalAs = MibScalar((1, 3, 6, 1, 2, 1, 15, 2), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 65535))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpLocalAs.setReference("RFC 4271, Section 4.2, 'My Autonomous System'.")
if mibBuilder.loadTexts: bgpLocalAs.setStatus('current')
if mibBuilder.loadTexts: bgpLocalAs.setDescription('The local autonomous system number.')
bgpPeerTable = MibTable((1, 3, 6, 1, 2, 1, 15, 3), )
if mibBuilder.loadTexts: bgpPeerTable.setStatus('current')
if mibBuilder.loadTexts: bgpPeerTable.setDescription('BGP peer table. This table contains, one entry per BGP peer, information about the connections with BGP peers.')
bgpPeerEntry = MibTableRow((1, 3, 6, 1, 2, 1, 15, 3, 1), ).setIndexNames((0, "BGP4-MIB", "bgpPeerRemoteAddr"))
if mibBuilder.loadTexts: bgpPeerEntry.setStatus('current')
if mibBuilder.loadTexts: bgpPeerEntry.setDescription('Entry containing information about the connection with a BGP peer.')
bgpPeerIdentifier = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 1), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerIdentifier.setReference("RFC 4271, Section 4.2, 'BGP Identifier'.")
if mibBuilder.loadTexts: bgpPeerIdentifier.setStatus('current')
if mibBuilder.loadTexts: bgpPeerIdentifier.setDescription("The BGP Identifier of this entry's BGP peer. This entry MUST be 0.0.0.0 unless the bgpPeerState is in the openconfirm or the established state.")
bgpPeerState = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 2), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3, 4, 5, 6))).clone(namedValues=NamedValues(("idle", 1), ("connect", 2), ("active", 3), ("opensent", 4), ("openconfirm", 5), ("established", 6)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerState.setReference('RFC 4271, Section 8.2.2.')
if mibBuilder.loadTexts: bgpPeerState.setStatus('current')
if mibBuilder.loadTexts: bgpPeerState.setDescription('The BGP peer connection state.')
bgpPeerAdminStatus = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 3), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2))).clone(namedValues=NamedValues(("stop", 1), ("start", 2)))).setMaxAccess("readwrite")
if mibBuilder.loadTexts: bgpPeerAdminStatus.setReference('RFC 4271, Section 8.1.2.')
if mibBuilder.loadTexts: bgpPeerAdminStatus.setStatus('current')
if mibBuilder.loadTexts: bgpPeerAdminStatus.setDescription("The desired state of the BGP connection. A transition from 'stop' to 'start' will cause the BGP Manual Start Event to be generated. A transition from 'start' to 'stop' will cause the BGP Manual Stop Event to be generated. This parameter can be used to restart BGP peer connections. Care should be used in providing write access to this object without adequate authentication.")
bgpPeerNegotiatedVersion = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 4), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerNegotiatedVersion.setReference('RFC 4271, Section 4.2. RFC 4271, Section 7.')
if mibBuilder.loadTexts: bgpPeerNegotiatedVersion.setStatus('current')
if mibBuilder.loadTexts: bgpPeerNegotiatedVersion.setDescription('The negotiated version of BGP running between the two peers. This entry MUST be zero (0) unless the bgpPeerState is in the openconfirm or the established state. Note that legal values for this object are between 0 and 255.')
bgpPeerLocalAddr = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 5), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerLocalAddr.setStatus('current')
if mibBuilder.loadTexts: bgpPeerLocalAddr.setDescription("The local IP address of this entry's BGP connection.")
bgpPeerLocalPort = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 6), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 65535))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerLocalPort.setStatus('current')
if mibBuilder.loadTexts: bgpPeerLocalPort.setDescription('The local port for the TCP connection between the BGP peers.')
bgpPeerRemoteAddr = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 7), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerRemoteAddr.setStatus('current')
if mibBuilder.loadTexts: bgpPeerRemoteAddr.setDescription("The remote IP address of this entry's BGP peer.")
bgpPeerRemotePort = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 8), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 65535))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerRemotePort.setStatus('current')
if mibBuilder.loadTexts: bgpPeerRemotePort.setDescription('The remote port for the TCP connection between the BGP peers. Note that the objects bgpPeerLocalAddr, bgpPeerLocalPort, bgpPeerRemoteAddr, and bgpPeerRemotePort provide the appropriate reference to the standard MIB TCP connection table.')
bgpPeerRemoteAs = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 9), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 65535))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerRemoteAs.setReference('RFC 4271, Section 4.2.')
if mibBuilder.loadTexts: bgpPeerRemoteAs.setStatus('current')
if mibBuilder.loadTexts: bgpPeerRemoteAs.setDescription('The remote autonomous system number received in the BGP OPEN message.')
bgpPeerInUpdates = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 10), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerInUpdates.setReference('RFC 4271, Section 4.3.')
if mibBuilder.loadTexts: bgpPeerInUpdates.setStatus('current')
if mibBuilder.loadTexts: bgpPeerInUpdates.setDescription('The number of BGP UPDATE messages received on this connection.')
bgpPeerOutUpdates = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 11), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerOutUpdates.setReference('RFC 4271, Section 4.3.')
if mibBuilder.loadTexts: bgpPeerOutUpdates.setStatus('current')
if mibBuilder.loadTexts: bgpPeerOutUpdates.setDescription('The number of BGP UPDATE messages transmitted on this connection.')
bgpPeerInTotalMessages = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 12), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerInTotalMessages.setReference('RFC 4271, Section 4.')
if mibBuilder.loadTexts: bgpPeerInTotalMessages.setStatus('current')
if mibBuilder.loadTexts: bgpPeerInTotalMessages.setDescription('The total number of messages received from the remote peer on this connection.')
bgpPeerOutTotalMessages = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 13), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerOutTotalMessages.setReference('RFC 4271, Section 4.')
if mibBuilder.loadTexts: bgpPeerOutTotalMessages.setStatus('current')
if mibBuilder.loadTexts: bgpPeerOutTotalMessages.setDescription('The total number of messages transmitted to the remote peer on this connection.')
bgpPeerLastError = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 14), OctetString().subtype(subtypeSpec=ValueSizeConstraint(2, 2)).setFixedLength(2)).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerLastError.setReference('RFC 4271, Section 4.5.')
if mibBuilder.loadTexts: bgpPeerLastError.setStatus('current')
if mibBuilder.loadTexts: bgpPeerLastError.setDescription('The last error code and subcode seen by this peer on this connection. If no error has occurred, this field is zero. Otherwise, the first byte of this two byte OCTET STRING contains the error code, and the second byte contains the subcode.')
bgpPeerFsmEstablishedTransitions = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 15), Counter32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerFsmEstablishedTransitions.setReference('RFC 4271, Section 8.')
if mibBuilder.loadTexts: bgpPeerFsmEstablishedTransitions.setStatus('current')
if mibBuilder.loadTexts: bgpPeerFsmEstablishedTransitions.setDescription('The total number of times the BGP FSM transitioned into the established state for this peer.')
bgpPeerFsmEstablishedTime = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 16), Gauge32()).setUnits('seconds').setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerFsmEstablishedTime.setReference('RFC 4271, Section 8.')
if mibBuilder.loadTexts: bgpPeerFsmEstablishedTime.setStatus('current')
if mibBuilder.loadTexts: bgpPeerFsmEstablishedTime.setDescription('This timer indicates how long (in seconds) this peer has been in the established state or how long since this peer was last in the established state. It is set to zero when a new peer is configured or when the router is booted.')
bgpPeerConnectRetryInterval = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 17), Integer32().subtype(subtypeSpec=ValueRangeConstraint(1, 65535))).setUnits('seconds').setMaxAccess("readwrite")
if mibBuilder.loadTexts: bgpPeerConnectRetryInterval.setReference("RFC 4271, Section 8.2.2. This is the value used to initialize the 'ConnectRetryTimer'.")
if mibBuilder.loadTexts: bgpPeerConnectRetryInterval.setStatus('current')
if mibBuilder.loadTexts: bgpPeerConnectRetryInterval.setDescription('Time interval (in seconds) for the ConnectRetry timer. The suggested value for this timer is 120 seconds.')
bgpPeerHoldTime = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 18), Integer32().subtype(subtypeSpec=ConstraintsUnion(ValueRangeConstraint(0, 0), ValueRangeConstraint(3, 65535), ))).setUnits('seconds').setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerHoldTime.setReference('RFC 4271, Section 4.2.')
if mibBuilder.loadTexts: bgpPeerHoldTime.setStatus('current')
if mibBuilder.loadTexts: bgpPeerHoldTime.setDescription('Time interval (in seconds) for the Hold Timer established with the peer. The value of this object is calculated by this BGP speaker, using the smaller of the values in bgpPeerHoldTimeConfigured and the Hold Time received in the OPEN message. This value must be at least three seconds if it is not zero (0). If the Hold Timer has not been established with the peer this object MUST have a value of zero (0). If the bgpPeerHoldTimeConfigured object has a value of (0), then this object MUST have a value of (0).')
bgpPeerKeepAlive = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 19), Integer32().subtype(subtypeSpec=ConstraintsUnion(ValueRangeConstraint(0, 0), ValueRangeConstraint(1, 21845), ))).setUnits('seconds').setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerKeepAlive.setReference('RFC 4271, Section 4.4.')
if mibBuilder.loadTexts: bgpPeerKeepAlive.setStatus('current')
if mibBuilder.loadTexts: bgpPeerKeepAlive.setDescription('Time interval (in seconds) for the KeepAlive timer established with the peer. The value of this object is calculated by this BGP speaker such that, when compared with bgpPeerHoldTime, it has the same proportion that bgpPeerKeepAliveConfigured has, compared with bgpPeerHoldTimeConfigured. If the KeepAlive timer has not been established with the peer, this object MUST have a value of zero (0). If the of bgpPeerKeepAliveConfigured object has a value of (0), then this object MUST have a value of (0).')
bgpPeerHoldTimeConfigured = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 20), Integer32().subtype(subtypeSpec=ConstraintsUnion(ValueRangeConstraint(0, 0), ValueRangeConstraint(3, 65535), ))).setUnits('seconds').setMaxAccess("readwrite")
if mibBuilder.loadTexts: bgpPeerHoldTimeConfigured.setReference('RFC 4271, Section 4.2. RFC 4271, Section 10.')
if mibBuilder.loadTexts: bgpPeerHoldTimeConfigured.setStatus('current')
if mibBuilder.loadTexts: bgpPeerHoldTimeConfigured.setDescription('Time interval (in seconds) for the Hold Time configured for this BGP speaker with this peer. This value is placed in an OPEN message sent to this peer by this BGP speaker, and is compared with the Hold Time field in an OPEN message received from the peer when determining the Hold Time (bgpPeerHoldTime) with the peer. This value must not be less than three seconds if it is not zero (0). If it is zero (0), the Hold Time is NOT to be established with the peer. The suggested value for this timer is 90 seconds.')
bgpPeerKeepAliveConfigured = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 21), Integer32().subtype(subtypeSpec=ConstraintsUnion(ValueRangeConstraint(0, 0), ValueRangeConstraint(1, 21845), ))).setUnits('seconds').setMaxAccess("readwrite")
if mibBuilder.loadTexts: bgpPeerKeepAliveConfigured.setReference('RFC 4271, Section 4.4. RFC 4271, Section 10.')
if mibBuilder.loadTexts: bgpPeerKeepAliveConfigured.setStatus('current')
if mibBuilder.loadTexts: bgpPeerKeepAliveConfigured.setDescription("Time interval (in seconds) for the KeepAlive timer configured for this BGP speaker with this peer. The value of this object will only determine the KEEPALIVE messages' frequency relative to the value specified in bgpPeerHoldTimeConfigured; the actual time interval for the KEEPALIVE messages is indicated by bgpPeerKeepAlive. A reasonable maximum value for this timer would be one third of that of bgpPeerHoldTimeConfigured. If the value of this object is zero (0), no periodical KEEPALIVE messages are sent to the peer after the BGP connection has been established. The suggested value for this timer is 30 seconds.")
bgpPeerMinASOriginationInterval = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 22), Integer32().subtype(subtypeSpec=ValueRangeConstraint(1, 65535))).setUnits('seconds').setMaxAccess("readwrite")
if mibBuilder.loadTexts: bgpPeerMinASOriginationInterval.setReference('RFC 4271, Section 9.2.1.2. RFC 4271, Section 10.')
if mibBuilder.loadTexts: bgpPeerMinASOriginationInterval.setStatus('current')
if mibBuilder.loadTexts: bgpPeerMinASOriginationInterval.setDescription('Time interval (in seconds) for the MinASOriginationInterval timer. The suggested value for this timer is 15 seconds.')
bgpPeerMinRouteAdvertisementInterval = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 23), Integer32().subtype(subtypeSpec=ValueRangeConstraint(1, 65535))).setUnits('seconds').setMaxAccess("readwrite")
if mibBuilder.loadTexts: bgpPeerMinRouteAdvertisementInterval.setReference('RFC 4271, Section 9.2.1.1. RFC 4271, Section 10.')
if mibBuilder.loadTexts: bgpPeerMinRouteAdvertisementInterval.setStatus('current')
if mibBuilder.loadTexts: bgpPeerMinRouteAdvertisementInterval.setDescription('Time interval (in seconds) for the MinRouteAdvertisementInterval timer. The suggested value for this timer is 30 seconds for EBGP connections and 5 seconds for IBGP connections.')
bgpPeerInUpdateElapsedTime = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 3, 1, 24), Gauge32()).setUnits('seconds').setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPeerInUpdateElapsedTime.setReference('RFC 4271, Section 4.3. RFC 4271, Section 8.2.2, Established state.')
if mibBuilder.loadTexts: bgpPeerInUpdateElapsedTime.setStatus('current')
if mibBuilder.loadTexts: bgpPeerInUpdateElapsedTime.setDescription('Elapsed time (in seconds) since the last BGP UPDATE message was received from the peer. Each time bgpPeerInUpdates is incremented, the value of this object is set to zero (0).')
bgpIdentifier = MibScalar((1, 3, 6, 1, 2, 1, 15, 4), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpIdentifier.setReference('RFC 4271, Section 4.2.')
if mibBuilder.loadTexts: bgpIdentifier.setStatus('current')
if mibBuilder.loadTexts: bgpIdentifier.setDescription('The BGP Identifier of the local system.')
bgpRcvdPathAttrTable = MibTable((1, 3, 6, 1, 2, 1, 15, 5), )
if mibBuilder.loadTexts: bgpRcvdPathAttrTable.setStatus('obsolete')
if mibBuilder.loadTexts: bgpRcvdPathAttrTable.setDescription('The BGP Received Path Attribute Table contains information about paths to destination networks, received from all peers running BGP version 3 or less.')
bgpPathAttrEntry = MibTableRow((1, 3, 6, 1, 2, 1, 15, 5, 1), ).setIndexNames((0, "BGP4-MIB", "bgpPathAttrDestNetwork"), (0, "BGP4-MIB", "bgpPathAttrPeer"))
if mibBuilder.loadTexts: bgpPathAttrEntry.setStatus('obsolete')
if mibBuilder.loadTexts: bgpPathAttrEntry.setDescription('Information about a path to a network.')
bgpPathAttrPeer = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 5, 1, 1), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPathAttrPeer.setStatus('obsolete')
if mibBuilder.loadTexts: bgpPathAttrPeer.setDescription('The IP address of the peer where the path information was learned.')
bgpPathAttrDestNetwork = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 5, 1, 2), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPathAttrDestNetwork.setReference('RFC 1267, Section 4.3.')
if mibBuilder.loadTexts: bgpPathAttrDestNetwork.setStatus('obsolete')
if mibBuilder.loadTexts: bgpPathAttrDestNetwork.setDescription('The address of the destination network.')
bgpPathAttrOrigin = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 5, 1, 3), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3))).clone(namedValues=NamedValues(("igp", 1), ("egp", 2), ("incomplete", 3)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPathAttrOrigin.setReference('RFC 1267, Section 4.3. RFC 1267, Section 5.')
if mibBuilder.loadTexts: bgpPathAttrOrigin.setStatus('obsolete')
if mibBuilder.loadTexts: bgpPathAttrOrigin.setDescription('The ultimate origin of the path information.')
bgpPathAttrASPath = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 5, 1, 4), OctetString().subtype(subtypeSpec=ValueSizeConstraint(2, 255))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPathAttrASPath.setReference('RFC 1267, Section 4.3. RFC 1267, Section 5.')
if mibBuilder.loadTexts: bgpPathAttrASPath.setStatus('obsolete')
if mibBuilder.loadTexts: bgpPathAttrASPath.setDescription('The set of ASes that must be traversed to reach the network. This object is probably best represented as SEQUENCE OF INTEGER. For SMI compatibility, though, it is represented as OCTET STRING. Each AS is represented as a pair of octets according to the following algorithm: first-byte-of-pair = ASNumber / 256; second-byte-of-pair = ASNumber & 255;')
bgpPathAttrNextHop = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 5, 1, 5), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPathAttrNextHop.setReference('RFC 1267, Section 4.3. RFC 1267, Section 5.')
if mibBuilder.loadTexts: bgpPathAttrNextHop.setStatus('obsolete')
if mibBuilder.loadTexts: bgpPathAttrNextHop.setDescription('The address of the border router that should be used for the destination network.')
bgpPathAttrInterASMetric = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 5, 1, 6), Integer32()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgpPathAttrInterASMetric.setReference('RFC 1267, Section 4.3. RFC 1267, Section 5.')
if mibBuilder.loadTexts: bgpPathAttrInterASMetric.setStatus('obsolete')
if mibBuilder.loadTexts: bgpPathAttrInterASMetric.setDescription('The optional inter-AS metric. If this attribute has not been provided for this route, the value for this object is 0.')
bgp4PathAttrTable = MibTable((1, 3, 6, 1, 2, 1, 15, 6), )
if mibBuilder.loadTexts: bgp4PathAttrTable.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrTable.setDescription('The BGP-4 Received Path Attribute Table contains information about paths to destination networks, received from all BGP4 peers.')
bgp4PathAttrEntry = MibTableRow((1, 3, 6, 1, 2, 1, 15, 6, 1), ).setIndexNames((0, "BGP4-MIB", "bgp4PathAttrIpAddrPrefix"), (0, "BGP4-MIB", "bgp4PathAttrIpAddrPrefixLen"), (0, "BGP4-MIB", "bgp4PathAttrPeer"))
if mibBuilder.loadTexts: bgp4PathAttrEntry.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrEntry.setDescription('Information about a path to a network.')
bgp4PathAttrPeer = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 1), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrPeer.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrPeer.setDescription('The IP address of the peer where the path information was learned.')
bgp4PathAttrIpAddrPrefixLen = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 2), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 32))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrIpAddrPrefixLen.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrIpAddrPrefixLen.setDescription('Length in bits of the IP address prefix in the Network Layer Reachability Information field.')
bgp4PathAttrIpAddrPrefix = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 3), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrIpAddrPrefix.setReference('RFC 4271, Section 4.3.')
if mibBuilder.loadTexts: bgp4PathAttrIpAddrPrefix.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrIpAddrPrefix.setDescription('An IP address prefix in the Network Layer Reachability Information field. This object is an IP address containing the prefix with length specified by bgp4PathAttrIpAddrPrefixLen. Any bits beyond the length specified by bgp4PathAttrIpAddrPrefixLen are zeroed.')
bgp4PathAttrOrigin = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 4), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2, 3))).clone(namedValues=NamedValues(("igp", 1), ("egp", 2), ("incomplete", 3)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrOrigin.setReference('RFC 4271, Section 4.3. RFC 4271, Section 5.1.1.')
if mibBuilder.loadTexts: bgp4PathAttrOrigin.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrOrigin.setDescription('The ultimate origin of the path information.')
bgp4PathAttrASPathSegment = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 5), OctetString().subtype(subtypeSpec=ValueSizeConstraint(2, 255))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrASPathSegment.setReference('RFC 4271, Section 4.3. RFC 4271, Section 5.1.2.')
if mibBuilder.loadTexts: bgp4PathAttrASPathSegment.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrASPathSegment.setDescription('The sequence of AS path segments. Each AS path segment is represented by a triple <type, length, value>. The type is a 1-octet field that has two possible values: 1 AS_SET: unordered set of ASes that a route in the UPDATE message has traversed 2 AS_SEQUENCE: ordered set of ASes that a route in the UPDATE message has traversed. The length is a 1-octet field containing the number of ASes in the value field. The value field contains one or more AS numbers. Each AS is represented in the octet string as a pair of octets according to the following algorithm: first-byte-of-pair = ASNumber / 256; second-byte-of-pair = ASNumber & 255; Known Issues: o BGP Confederations will result in a type of either 3 or 4. o An AS Path may be longer than 255 octets. This may result in this object containing a truncated AS Path.')
bgp4PathAttrNextHop = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 6), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrNextHop.setReference('RFC 4271, Section 4.3. RFC 4271, Section 5.1.3.')
if mibBuilder.loadTexts: bgp4PathAttrNextHop.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrNextHop.setDescription('The address of the border router that should be used for the destination network. This address is the NEXT_HOP address received in the UPDATE packet.')
bgp4PathAttrMultiExitDisc = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 7), Integer32().subtype(subtypeSpec=ValueRangeConstraint(-1, 2147483647))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrMultiExitDisc.setReference('RFC 4271, Section 4.3. RFC 4271, Section 5.1.4.')
if mibBuilder.loadTexts: bgp4PathAttrMultiExitDisc.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrMultiExitDisc.setDescription('This metric is used to discriminate between multiple exit points to an adjacent autonomous system. A value of -1 indicates the absence of this attribute. Known Issues: o The BGP-4 specification uses an unsigned 32 bit number. Thus, this object cannot represent the full range of the protocol.')
bgp4PathAttrLocalPref = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 8), Integer32().subtype(subtypeSpec=ValueRangeConstraint(-1, 2147483647))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrLocalPref.setReference('RFC 4271, Section 4.3. RFC 4271, Section 5.1.5.')
if mibBuilder.loadTexts: bgp4PathAttrLocalPref.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrLocalPref.setDescription("The originating BGP4 speaker's degree of preference for an advertised route. A value of -1 indicates the absence of this attribute. Known Issues: o The BGP-4 specification uses an unsigned 32 bit number and thus this object cannot represent the full range of the protocol.")
bgp4PathAttrAtomicAggregate = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 9), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2))).clone(namedValues=NamedValues(("lessSpecificRouteNotSelected", 1), ("lessSpecificRouteSelected", 2)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrAtomicAggregate.setReference('RFC 4271, Sections 5.1.6 and 9.1.4.')
if mibBuilder.loadTexts: bgp4PathAttrAtomicAggregate.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrAtomicAggregate.setDescription("If the ATOMIC_AGGREGATE attribute is present in the Path Attributes then this object MUST have a value of 'lessSpecificRouteNotSelected'. If the ATOMIC_AGGREGATE attribute is missing in the Path Attributes then this object MUST have a value of 'lessSpecificRouteSelected'. Note that ATOMIC_AGGREGATE is now a primarily informational attribute.")
bgp4PathAttrAggregatorAS = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 10), Integer32().subtype(subtypeSpec=ValueRangeConstraint(0, 65535))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrAggregatorAS.setReference('RFC 4271, Section 5.1.7. RFC 4271, Section 9.2.2.2.')
if mibBuilder.loadTexts: bgp4PathAttrAggregatorAS.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrAggregatorAS.setDescription('The AS number of the last BGP4 speaker that performed route aggregation. A value of zero (0) indicates the absence of this attribute. Note that propagation of AS of zero is illegal in the Internet.')
bgp4PathAttrAggregatorAddr = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 11), IpAddress()).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrAggregatorAddr.setReference('RFC 4271, Section 5.1.7. RFC 4271, Section 9.2.2.2.')
if mibBuilder.loadTexts: bgp4PathAttrAggregatorAddr.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrAggregatorAddr.setDescription('The IP address of the last BGP4 speaker that performed route aggregation. A value of 0.0.0.0 indicates the absence of this attribute.')
bgp4PathAttrCalcLocalPref = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 12), Integer32().subtype(subtypeSpec=ValueRangeConstraint(-1, 2147483647))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrCalcLocalPref.setReference('RFC 4271, Section 9.1.1.')
if mibBuilder.loadTexts: bgp4PathAttrCalcLocalPref.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrCalcLocalPref.setDescription('The degree of preference calculated by the receiving BGP4 speaker for an advertised route. A value of -1 indicates the absence of this attribute. Known Issues: o The BGP-4 specification uses an unsigned 32 bit number and thus this object cannot represent the full range of the protocol.')
bgp4PathAttrBest = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 13), Integer32().subtype(subtypeSpec=ConstraintsUnion(SingleValueConstraint(1, 2))).clone(namedValues=NamedValues(("false", 1), ("true", 2)))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrBest.setReference('RFC 4271, Section 9.1.2.')
if mibBuilder.loadTexts: bgp4PathAttrBest.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrBest.setDescription('An indication of whether this route was chosen as the best BGP4 route for this destination.')
bgp4PathAttrUnknown = MibTableColumn((1, 3, 6, 1, 2, 1, 15, 6, 1, 14), OctetString().subtype(subtypeSpec=ValueSizeConstraint(0, 255))).setMaxAccess("readonly")
if mibBuilder.loadTexts: bgp4PathAttrUnknown.setReference('RFC 4271, Section 5.')
if mibBuilder.loadTexts: bgp4PathAttrUnknown.setStatus('current')
if mibBuilder.loadTexts: bgp4PathAttrUnknown.setDescription('One or more path attributes not understood by this BGP4 speaker. Path attributes are recorded in the Update Path attribute format of type, length, value. Size zero (0) indicates the absence of such attributes. Octets beyond the maximum size, if any, are not recorded by this object. Known Issues: o Attributes understood by this speaker, but not represented in this MIB, are unavailable to the agent.')
bgpNotification = MibIdentifier((1, 3, 6, 1, 2, 1, 15, 0))
bgpEstablishedNotification = NotificationType((1, 3, 6, 1, 2, 1, 15, 0, 1)).setObjects(("BGP4-MIB", "bgpPeerRemoteAddr"), ("BGP4-MIB", "bgpPeerLastError"), ("BGP4-MIB", "bgpPeerState"))
if mibBuilder.loadTexts: bgpEstablishedNotification.setStatus('current')
if mibBuilder.loadTexts: bgpEstablishedNotification.setDescription('The bgpEstablishedNotification event is generated when the BGP FSM enters the established state. This Notification replaces the bgpEstablished Notification.')
bgpBackwardTransNotification = NotificationType((1, 3, 6, 1, 2, 1, 15, 0, 2)).setObjects(("BGP4-MIB", "bgpPeerRemoteAddr"), ("BGP4-MIB", "bgpPeerLastError"), ("BGP4-MIB", "bgpPeerState"))
if mibBuilder.loadTexts: bgpBackwardTransNotification.setStatus('current')
if mibBuilder.loadTexts: bgpBackwardTransNotification.setDescription('The bgpBackwardTransNotification event is generated when the BGP FSM moves from a higher numbered state to a lower numbered state. This Notification replaces the bgpBackwardsTransition Notification.')
bgpTraps = MibIdentifier((1, 3, 6, 1, 2, 1, 15, 7))
bgpEstablished = NotificationType((1, 3, 6, 1, 2, 1, 15, 7, 1)).setObjects(("BGP4-MIB", "bgpPeerLastError"), ("BGP4-MIB", "bgpPeerState"))
if mibBuilder.loadTexts: bgpEstablished.setStatus('deprecated')
if mibBuilder.loadTexts: bgpEstablished.setDescription('The bgpEstablished event is generated when the BGP FSM enters the established state. This Notification has been replaced by the bgpEstablishedNotification Notification.')
bgpBackwardTransition = NotificationType((1, 3, 6, 1, 2, 1, 15, 7, 2)).setObjects(("BGP4-MIB", "bgpPeerLastError"), ("BGP4-MIB", "bgpPeerState"))
if mibBuilder.loadTexts: bgpBackwardTransition.setStatus('deprecated')
if mibBuilder.loadTexts: bgpBackwardTransition.setDescription('The bgpBackwardTransition event is generated when the BGP FSM moves from a higher numbered state to a lower numbered state. This Notification has been replaced by the bgpBackwardTransNotification Notification.')
bgp4MIBConformance = MibIdentifier((1, 3, 6, 1, 2, 1, 15, 8))
bgp4MIBCompliances = MibIdentifier((1, 3, 6, 1, 2, 1, 15, 8, 1))
bgp4MIBGroups = MibIdentifier((1, 3, 6, 1, 2, 1, 15, 8, 2))
bgp4MIBCompliance = ModuleCompliance((1, 3, 6, 1, 2, 1, 15, 8, 1, 1)).setObjects(("BGP4-MIB", "bgp4MIBGlobalsGroup"), ("BGP4-MIB", "bgp4MIBPeerGroup"), ("BGP4-MIB", "bgp4MIBPathAttrGroup"), ("BGP4-MIB", "bgp4MIBNotificationGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bgp4MIBCompliance = bgp4MIBCompliance.setStatus('current')
if mibBuilder.loadTexts: bgp4MIBCompliance.setDescription('The compliance statement for entities which implement the BGP4 mib.')
bgp4MIBDeprecatedCompliances = ModuleCompliance((1, 3, 6, 1, 2, 1, 15, 8, 1, 2)).setObjects(("BGP4-MIB", "bgp4MIBTrapGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bgp4MIBDeprecatedCompliances = bgp4MIBDeprecatedCompliances.setStatus('deprecated')
if mibBuilder.loadTexts: bgp4MIBDeprecatedCompliances.setDescription('The compliance statement documenting deprecated objects in the BGP4 mib.')
bgp4MIBObsoleteCompliances = ModuleCompliance((1, 3, 6, 1, 2, 1, 15, 8, 1, 3)).setObjects(("BGP4-MIB", "bgpRcvdPathAttrGroup"))

if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bgp4MIBObsoleteCompliances = bgp4MIBObsoleteCompliances.setStatus('obsolete')
if mibBuilder.loadTexts: bgp4MIBObsoleteCompliances.setDescription('The compliance statement documenting obsolete objects in the BGP4 mib.')
bgp4MIBGlobalsGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 15, 8, 2, 1)).setObjects(("BGP4-MIB", "bgpVersion"), ("BGP4-MIB", "bgpLocalAs"), ("BGP4-MIB", "bgpIdentifier"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bgp4MIBGlobalsGroup = bgp4MIBGlobalsGroup.setStatus('current')
if mibBuilder.loadTexts: bgp4MIBGlobalsGroup.setDescription('A collection of objects providing information on global BGP state.')
bgp4MIBPeerGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 15, 8, 2, 2)).setObjects(("BGP4-MIB", "bgpPeerIdentifier"), ("BGP4-MIB", "bgpPeerState"), ("BGP4-MIB", "bgpPeerAdminStatus"), ("BGP4-MIB", "bgpPeerNegotiatedVersion"), ("BGP4-MIB", "bgpPeerLocalAddr"), ("BGP4-MIB", "bgpPeerLocalPort"), ("BGP4-MIB", "bgpPeerRemoteAddr"), ("BGP4-MIB", "bgpPeerRemotePort"), ("BGP4-MIB", "bgpPeerRemoteAs"), ("BGP4-MIB", "bgpPeerInUpdates"), ("BGP4-MIB", "bgpPeerOutUpdates"), ("BGP4-MIB", "bgpPeerInTotalMessages"), ("BGP4-MIB", "bgpPeerOutTotalMessages"), ("BGP4-MIB", "bgpPeerLastError"), ("BGP4-MIB", "bgpPeerFsmEstablishedTransitions"), ("BGP4-MIB", "bgpPeerFsmEstablishedTime"), ("BGP4-MIB", "bgpPeerConnectRetryInterval"), ("BGP4-MIB", "bgpPeerHoldTime"), ("BGP4-MIB", "bgpPeerKeepAlive"), ("BGP4-MIB", "bgpPeerHoldTimeConfigured"), ("BGP4-MIB", "bgpPeerKeepAliveConfigured"), ("BGP4-MIB", "bgpPeerMinASOriginationInterval"), ("BGP4-MIB", "bgpPeerMinRouteAdvertisementInterval"), ("BGP4-MIB", "bgpPeerInUpdateElapsedTime"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bgp4MIBPeerGroup = bgp4MIBPeerGroup.setStatus('current')
if mibBuilder.loadTexts: bgp4MIBPeerGroup.setDescription('A collection of objects for managing BGP peers.')
bgpRcvdPathAttrGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 15, 8, 2, 3)).setObjects(("BGP4-MIB", "bgpPathAttrPeer"), ("BGP4-MIB", "bgpPathAttrDestNetwork"), ("BGP4-MIB", "bgpPathAttrOrigin"), ("BGP4-MIB", "bgpPathAttrASPath"), ("BGP4-MIB", "bgpPathAttrNextHop"), ("BGP4-MIB", "bgpPathAttrInterASMetric"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bgpRcvdPathAttrGroup = bgpRcvdPathAttrGroup.setStatus('obsolete')
if mibBuilder.loadTexts: bgpRcvdPathAttrGroup.setDescription('A collection of objects for managing BGP-3 and earlier path entries. This conformance group, like BGP-3, is obsolete.')
bgp4MIBPathAttrGroup = ObjectGroup((1, 3, 6, 1, 2, 1, 15, 8, 2, 4)).setObjects(("BGP4-MIB", "bgp4PathAttrPeer"), ("BGP4-MIB", "bgp4PathAttrIpAddrPrefixLen"), ("BGP4-MIB", "bgp4PathAttrIpAddrPrefix"), ("BGP4-MIB", "bgp4PathAttrOrigin"), ("BGP4-MIB", "bgp4PathAttrASPathSegment"), ("BGP4-MIB", "bgp4PathAttrNextHop"), ("BGP4-MIB", "bgp4PathAttrMultiExitDisc"), ("BGP4-MIB", "bgp4PathAttrLocalPref"), ("BGP4-MIB", "bgp4PathAttrAtomicAggregate"), ("BGP4-MIB", "bgp4PathAttrAggregatorAS"), ("BGP4-MIB", "bgp4PathAttrAggregatorAddr"), ("BGP4-MIB", "bgp4PathAttrCalcLocalPref"), ("BGP4-MIB", "bgp4PathAttrBest"), ("BGP4-MIB", "bgp4PathAttrUnknown"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bgp4MIBPathAttrGroup = bgp4MIBPathAttrGroup.setStatus('current')
if mibBuilder.loadTexts: bgp4MIBPathAttrGroup.setDescription('A collection of objects for managing BGP path entries.')
bgp4MIBTrapGroup = NotificationGroup((1, 3, 6, 1, 2, 1, 15, 8, 2, 5)).setObjects(("BGP4-MIB", "bgpEstablished"), ("BGP4-MIB", "bgpBackwardTransition"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bgp4MIBTrapGroup = bgp4MIBTrapGroup.setStatus('deprecated')
if mibBuilder.loadTexts: bgp4MIBTrapGroup.setDescription('A collection of notifications for signaling changes in BGP peer relationships. Obsoleted by bgp4MIBNotificationGroup')
bgp4MIBNotificationGroup = NotificationGroup((1, 3, 6, 1, 2, 1, 15, 8, 2, 6)).setObjects(("BGP4-MIB", "bgpEstablishedNotification"), ("BGP4-MIB", "bgpBackwardTransNotification"))
if getattr(mibBuilder, 'version', (0, 0, 0)) > (4, 4, 0):
    bgp4MIBNotificationGroup = bgp4MIBNotificationGroup.setStatus('current')
if mibBuilder.loadTexts: bgp4MIBNotificationGroup.setDescription('A collection of notifications for signaling changes in BGP peer relationships. Obsoletes bgp4MIBTrapGroup.')
mibBuilder.exportSymbols("BGP4-MIB", bgp4PathAttrEntry=bgp4PathAttrEntry, bgp4PathAttrLocalPref=bgp4PathAttrLocalPref, bgpTraps=bgpTraps, bgpPeerHoldTimeConfigured=bgpPeerHoldTimeConfigured, bgpPeerInUpdates=bgpPeerInUpdates, bgpBackwardTransNotification=bgpBackwardTransNotification, bgpPeerEntry=bgpPeerEntry, bgpPeerLastError=bgpPeerLastError, bgpPathAttrPeer=bgpPathAttrPeer, bgp4MIBGlobalsGroup=bgp4MIBGlobalsGroup, bgp4MIBNotificationGroup=bgp4MIBNotificationGroup, bgpBackwardTransition=bgpBackwardTransition, bgpPeerRemotePort=bgpPeerRemotePort, bgp4MIBPeerGroup=bgp4MIBPeerGroup, bgpRcvdPathAttrGroup=bgpRcvdPathAttrGroup, bgpPeerConnectRetryInterval=bgpPeerConnectRetryInterval, bgpPeerMinASOriginationInterval=bgpPeerMinASOriginationInterval, bgpPeerIdentifier=bgpPeerIdentifier, bgpPathAttrNextHop=bgpPathAttrNextHop, bgp4PathAttrAggregatorAS=bgp4PathAttrAggregatorAS, bgpPathAttrDestNetwork=bgpPathAttrDestNetwork, bgp4PathAttrIpAddrPrefixLen=bgp4PathAttrIpAddrPrefixLen, bgpIdentifier=bgpIdentifier, bgp4MIBObsoleteCompliances=bgp4MIBObsoleteCompliances, bgp=bgp, bgpPeerInTotalMessages=bgpPeerInTotalMessages, bgpPeerInUpdateElapsedTime=bgpPeerInUpdateElapsedTime, bgpPathAttrInterASMetric=bgpPathAttrInterASMetric, bgp4PathAttrTable=bgp4PathAttrTable, bgp4PathAttrNextHop=bgp4PathAttrNextHop, bgp4PathAttrCalcLocalPref=bgp4PathAttrCalcLocalPref, bgpPeerTable=bgpPeerTable, bgpVersion=bgpVersion, bgp4PathAttrIpAddrPrefix=bgp4PathAttrIpAddrPrefix, bgp4MIBGroups=bgp4MIBGroups, bgp4PathAttrOrigin=bgp4PathAttrOrigin, bgp4MIBTrapGroup=bgp4MIBTrapGroup, bgpRcvdPathAttrTable=bgpRcvdPathAttrTable, bgpPeerHoldTime=bgpPeerHoldTime, bgp4PathAttrASPathSegment=bgp4PathAttrASPathSegment, bgpEstablishedNotification=bgpEstablishedNotification, bgpPathAttrASPath=bgpPathAttrASPath, bgpPeerAdminStatus=bgpPeerAdminStatus, bgpPeerOutTotalMessages=bgpPeerOutTotalMessages, bgpPeerFsmEstablishedTransitions=bgpPeerFsmEstablishedTransitions, bgp4PathAttrAtomicAggregate=bgp4PathAttrAtomicAggregate, bgpPeerMinRouteAdvertisementInterval=bgpPeerMinRouteAdvertisementInterval, bgpEstablished=bgpEstablished, bgpNotification=bgpNotification, bgpPeerFsmEstablishedTime=bgpPeerFsmEstablishedTime, bgp4PathAttrAggregatorAddr=bgp4PathAttrAggregatorAddr, bgp4PathAttrBest=bgp4PathAttrBest, bgp4PathAttrPeer=bgp4PathAttrPeer, bgpPathAttrOrigin=bgpPathAttrOrigin, bgpPeerState=bgpPeerState, bgpPeerRemoteAs=bgpPeerRemoteAs, bgpPeerKeepAlive=bgpPeerKeepAlive, bgp4PathAttrMultiExitDisc=bgp4PathAttrMultiExitDisc, bgp4MIBCompliance=bgp4MIBCompliance, bgp4MIBPathAttrGroup=bgp4MIBPathAttrGroup, bgpPeerLocalAddr=bgpPeerLocalAddr, bgpPeerOutUpdates=bgpPeerOutUpdates, bgp4PathAttrUnknown=bgp4PathAttrUnknown, bgp4MIBDeprecatedCompliances=bgp4MIBDeprecatedCompliances, bgp4MIBConformance=bgp4MIBConformance, bgp4MIBCompliances=bgp4MIBCompliances, bgpPeerRemoteAddr=bgpPeerRemoteAddr, bgpPeerNegotiatedVersion=bgpPeerNegotiatedVersion, bgpPeerLocalPort=bgpPeerLocalPort, bgpPathAttrEntry=bgpPathAttrEntry, bgpPeerKeepAliveConfigured=bgpPeerKeepAliveConfigured, bgpLocalAs=bgpLocalAs, PYSNMP_MODULE_ID=bgp)



#!/usr/bin/env python3
"""Test SNMP polling using the PySNMP high-level API directly"""

import argparse
import asyncio
import logging

from pysnmp.hlapi.asyncio import (
    CommunityData,
    ContextData,
    ObjectIdentity,
    ObjectType,
    SnmpEngine,
    UdpTransportTarget,
    bulkCmd,
    isEndOfMib,
)

_log = logging.getLogger("polltest")


def main():
    args = parse_args()
    logging.basicConfig(
        level=logging.INFO, format="%(asctime)s - %(levelname)s - %(name)s (%(threadName)s) - %(message)s"
    )
    asyncio.run(run(args))


async def run(args: argparse.Namespace):
    snmp_engine = SnmpEngine()
    root = ObjectIdentity("IF-MIB", "ifDescr")
    query = [ObjectType(root)]
    while True:
        error_indication, error_status, error_index, var_binds = await bulkCmd(
            snmp_engine,
            CommunityData(args.community, mpModel=1),
            UdpTransportTarget((args.agent, 161)),
            ContextData(),
            0,
            50,
            *query,
            lexicographicMode=False,
        )

        if error_indication:
            _log.error(error_indication)
            break

        elif error_status:
            _log.error("%s at %s", error_status.prettyPrint(), error_index and query[int(error_index) - 1][0] or "?")
        else:
            for row in var_binds:
                for var_bind in row:
                    # this break is only necessary if the query is translated to a get-next command on a v1 session
                    if not root.getOid().isPrefixOf(var_bind[0]):
                        break
                    _log.info(" = ".join([x.prettyPrint() for x in var_bind]))
        query = var_binds[-1]
        # this break is only necessary if the query is translated to a get-next command on a v1 session
        if not root.getOid().isPrefixOf(query[0][0]):
            break

        if isEndOfMib(query):
            break

    snmp_engine.transportDispatcher.closeDispatcher()


def parse_args():
    parser = argparse.ArgumentParser(description="Fetch ifDescr values from an SNMP agent")
    parser.add_argument("agent", type=str, help="SNMP agent name/address")
    parser.add_argument(
        "-c",
        "--community",
        type=str,
        metavar="COMMUNITY",
        default="public",
        help="SNMP community to use (if not `public`)",
    )
    return parser.parse_args()


if __name__ == "__main__":
    main()



import logging
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Dict, Optional, Protocol, Union

from pydantic.main import BaseModel

from zino.statemodels import (
    DeviceMaintenance,
    Event,
    EventState,
    MatchType,
    PlannedMaintenance,
    PortStateMaintenance,
)
from zino.time import now as utcnow

if TYPE_CHECKING:
    from zino.state import ZinoState

_log = logging.getLogger(__name__)


PM_EXPIRY_TIME = timedelta(days=3)


class PlannedMaintenanceObserver(Protocol):
    """Defines a valid protocol for planned maintenance observer functions"""

    def __call__(self) -> None: ...


class PlannedMaintenances(BaseModel):
    planned_maintenances: Dict[int, Union[DeviceMaintenance, PortStateMaintenance]] = {}
    last_pm_id: int = 0
    last_run: Optional[datetime] = datetime.fromtimestamp(0, tz=timezone.utc)
    _observers: list[PlannedMaintenanceObserver] = []

    def __getitem__(self, item):
        return self.planned_maintenances[item]

    def __len__(self):
        return len(self.planned_maintenances)

    def create_planned_maintenance(
        self,
        start_time: datetime,
        end_time: datetime,
        pm_class: type[PlannedMaintenance],
        match_type: MatchType,
        match_expression: str,
        match_device: Optional[str] = None,
    ) -> PlannedMaintenance:
        """Creates a planned maintenance, adds it to the planned_maintenances dict and
        returns it
        """
        pm_id = self.get_next_available_pm_id()
        pm = pm_class(
            id=pm_id,
            start_time=start_time,
            end_time=end_time,
            match_type=match_type,
            match_expression=match_expression,
            match_device=match_device,
        )
        self.planned_maintenances[pm_id] = pm
        self._call_observers()
        return pm

    def close_planned_maintenance(self, id: int, reason: str, user: str) -> None:
        """Deletes planned maintenance with the given id"""
        pm = self.planned_maintenances.get(id, None)
        if not pm:
            return
        pm.add_log(f"PM closed by {user}: {reason}")
        del self.planned_maintenances[id]
        self._call_observers()

    def get_started_planned_maintenances(self, now: datetime) -> list[PlannedMaintenance]:
        """Returns all planned maintenances that have begun since the last run of this
        task until `now`
        """
        return [pm for pm in self.planned_maintenances.values() if self.last_run < pm.start_time <= now < pm.end_time]

    def get_ended_planned_maintenances(self, now: datetime) -> list[PlannedMaintenance]:
        """Returns all planned maintenances that have ended since the last run of this
        task until `now`
        """
        return [pm for pm in self.planned_maintenances.values() if self.last_run < pm.end_time <= now]

    def get_active_planned_maintenances(self, now: datetime) -> list[PlannedMaintenance]:
        """Returns all planned maintenances that are currently active

        This means it has started before `now` and the end_time is later than `now`
        """
        return [pm for pm in self.planned_maintenances.values() if pm.start_time < now < pm.end_time]

    def get_old_planned_maintenances(self, now: datetime) -> list[PlannedMaintenance]:
        """Returns all planned maintenances that should get deleted

        This means that `now` is 3 days later than end_time
        """
        return [pm for pm in self.planned_maintenances.values() if now - pm.end_time > PM_EXPIRY_TIME]

    def add_pm_observer(self, observer: PlannedMaintenanceObserver) -> None:
        """Adds an observer function that will be called any time a planned maintenance
        is added or removed
        """
        self._observers.append(observer)

    def get_next_available_pm_id(self) -> int:
        """Returns the next available planned maintenance id"""
        self.last_pm_id += 1
        return self.last_pm_id

    def _call_observers(self) -> None:
        for observer in self._observers:
            observer()

    def update_pm_states(self, state: "ZinoState"):
        """This function starts and stops planned maintenances according to their
        schedule and updates affected events
        """
        now = utcnow()

        # Initiate PM once it becomes active
        for started_pm in self.get_started_planned_maintenances(now=now):
            started_pm.start(state)

        # Make sure all events that match a PM are ignored
        active_pms = self.get_active_planned_maintenances(now)
        for event in state.events.events.values():
            if event.state not in [EventState.IGNORED, EventState.CLOSED]:
                self._ignore_event_if_it_has_active_planned_maintenance(state, event, active_pms)

        # Set events matching ended PMs to open
        for ended_pm in self.get_ended_planned_maintenances(now=now):
            ended_pm.end(state)

        # Delete PMs that ended a certain amount of time ago
        old_pms = self.get_old_planned_maintenances(now)
        for pm in old_pms:
            self.close_planned_maintenance(pm.id, "timer expiry for old PMs", "zino")

        self.last_run = now

    def _ignore_event_if_it_has_active_planned_maintenance(
        self, state: "ZinoState", event: Event, active_pms: list[PlannedMaintenance]
    ):
        for pm in active_pms:
            if pm.matches_event(event, state):
                event.state = EventState.IGNORED
                event.add_log(f"entered into existing active PM event id {pm.id}")
                state.events.commit(event)
                if event.id not in pm.event_ids:
                    pm.event_ids.append(event.id)
